(* ProofBase contains definitions and proofs common to both the
   state-to-trace and trace-to-state semantics proofs,
   including:
    - initial
    - valid
    - Program_to_State
    - exact
    - eq_set
    - Label_to_newState
    - down_to_pre
    - eval_s1 *)



Require Import List.
Require Import SemanticsBase.
Require Import SemanticsState.
Require Import SemanticsTrace.
Require Import ProofPre.



(* ~~~~~~~ INITIAL ~~~~~~~ *)

Fixpoint initial (P : Program) : Prop :=
  match P with
    | skip => True
    | comm (bang A) => True
    | comm (bang' A) => False
    | comm (quer A) => True
    | comm (quer' A) => False
    | sequ P0 P1 => initial P0 /\ initial P1
    | par P0 P1 => initial P0 /\ initial P1
    | rep P' => initial P'
    | det GP0 GP1 => initial (snd GP0) /\ initial (snd GP1)
    | ndet GP0 GP1 => initial (snd GP0) /\ initial (snd GP1)
  end.



(* ~~~~~~~ VALID ~~~~~~~ *)

Fixpoint valid (P : Program) : Prop :=
  match P with
    | skip => True
    | comm _ => True
    | sequ P0 P1 => valid P0 /\ initial P1
    | par P0 P1 => valid P0 /\ valid P1
    | rep P' => initial P'
    | det GP0 GP1 => initial (snd GP0) /\ initial (snd GP1)
    | ndet GP0 GP1 => initial (snd GP0) /\ initial (snd GP1)
  end.

Fixpoint initial_to_valid {P} (i : initial P) : valid P :=
  match P as P return initial P -> valid P with
    | skip => fun _ => I
    | comm _ => fun _ => I
    | sequ _ _ => fun i => conj (initial_to_valid (proj1 i)) (proj2 i)
    | par _ _ => fun i => conj (initial_to_valid (proj1 i)) (initial_to_valid (proj2 i))
    | rep _ => fun i => i
    | det _ _ => fun i => i
    | ndet _ _ => fun i => i
  end i.



(* ~~~~~~~ EQ_STATE ~~~~~~~ *)

Definition eq_State_refl {s} : eq_State s s :=
  fun p => eq_refl _.

Definition eq_State_sym {s s'} (e : eq_State s s') : eq_State s' s :=
  fun p => eq_sym (e p).

Definition eq_State_trans {s s' s''} (e : eq_State s s') (e' : eq_State s' s'') : eq_State s s'' :=
  fun p => eq_trans (e p) (e' p).



(* ~~~~~~~ PROGRAM_TO_STATE ~~~~~~~ *)

Fixpoint Program_to_State (P : Program) : State :=
  match P with
    | comm (bang' A) => (bar A) :: nil
    | comm (quer' A) => (hat A) :: nil
    | sequ P0 P1 => Program_to_State P0
    | par P0 P1 => Program_to_State P0 ++ Program_to_State P1
    | _ => nil
  end.

Fixpoint initial_to_nil {P} (i : initial P) : Program_to_State P = nil :=
  match P as P return initial P -> Program_to_State P = nil with
    | comm (bang' A) => fun i => match i with end
    | comm (quer' A) => fun i => match i with end
    | sequ _ _ => fun i => initial_to_nil (proj1 i)
    | par _ _ => fun i => eq_ind_r (fun x => _ ++ x = _)
                                   (eq_ind_r (fun x => x = _)
                                             (eq_refl _)
                                             (eq_trans (app_nil_r _)
                                                       (initial_to_nil (proj1 i))))
                                   (initial_to_nil (proj2 i))
    | _ => fun _ => eq_refl _
  end i.



(* ~~~~~~~ EXACT ~~~~~~~ *)

Definition exact (P : Program) (s : State) : Prop :=
  eq_State (Program_to_State P) s.

Definition initial_to_exact {P} (i : initial P) : exact P state_init :=
  eq_ind_r (fun x => eq_State x _) eq_State_refl (initial_to_nil i).



(* ~~~~~~~ EQ_SET ~~~~~~~ *)

Definition eq_set (l l' : list Probe) : Prop :=
  forall p, In p l <-> In p l'.

Definition eq_State_to_eq_set {s s'} (es : eq_State s s') : eq_set s s' :=
  fun p => conj (fun is => (proj2 (count_occ_In _ _ _) (eq_ind_r (fun x => x > _)
                                                       (proj1 (count_occ_In _ _ _) is)
                                                       (eq_sym (es p)))))
                (fun is' => (proj2 (count_occ_In _ _ _) (eq_ind_r (fun x => x > _)
                                                        (proj1 (count_occ_In _ _ _) is')
                                                        (es p)))).

Fixpoint req_eq_set {a l l'} (r : req a l) (e : eq_set l l') : req a l' :=
  let F := fix f g := match g as g return req (holds g) l -> req (holds g) l' with
                        | nil => fun _ => I
                        | a :: g' => fun r => conj (proj1 (e _) (proj1 r)) (f g' (proj2 r))
                      end
  in match a as a return req a l -> req a l' with
       | idle => fun _ => I
       | step (send start A) => fun r => fun il' => r (proj2 (e _) il')
       | step (recv start A) => fun r => (proj1 (e _)) r
       | step (send use A) => fun _ => I
       | step (recv use A) => fun _ => I
       | step (send finish A) => fun r => (proj1 (e _)) r
       | step (recv finish A) => fun r => fun il' => r (proj2 (e _) il')
       | holds G => fun r => F G r
       | holds2 G1 G2 => fun r => conj (F G1 (proj1 r)) (F G2 (proj2 r))
       | ll a0 a1 => fun r => conj (req_eq_set (proj1 r) e) (req_eq_set (proj2 r) e)
     end r.



(* ~~~~~~~ LABEL_TO_NEWSTATE ~~~~~~~ *)

Fixpoint Label_to_newState (a : Label) (s : State) : list Probe :=
  match a with
    | step (send start A) => (bar A) :: s
    | step (send finish A) => remove eq_Probe_dec (bar A) s
    | step (recv start A) => (hat A) :: s
    | step (recv finish A) => remove eq_Probe_dec (hat A) s
    | ll a0 a1 => Label_to_newState a1 (Label_to_newState a0 s)
    | _ => s
  end.



(* ~~~~~~~ DOWN_TO_PRE ~~~~~~~ *)

Fixpoint down_to_pre {a p} (i : In p (down a)) : In p (pre a) :=
  match a as a return In p (down a) -> In p (pre a) with
    | step (send finish A) => fun i =>
                                match eq_Probe_dec p (bar A) with
                                  | left e => eq_ind_r (fun x => In x _) (in_eq _ _) e
                                  | right ne => match not_in_neq ne i with end end
    | step (recv finish A) => fun i =>
                                match eq_Probe_dec p (hat A) with
                                  | left e => eq_ind_r (fun x => In x _) (in_eq _ _) e
                                  | right ne => match not_in_neq ne i with end end
    | ll a0 a1 => fun i =>
                  match in_app_or _ _ _ i with
                    | or_introl i0 => in_or_app _ _ _ (or_introl (down_to_pre i0))
                    | or_intror i1 => in_or_app _ _ _ (or_intror (down_to_pre i1)) end
    | _ => fun i => match in_nil i with end
  end i.



(* ~~~~~~~ EVAL_S1 ~~~~~~~ *)

Fixpoint eval_s1 {x} (G : Guard) : eval G x <-> s1 x G :=
  match G as G return eval G x <-> s1 x G with
    | nil => conj (fun e => e) (fun s => s)
    | A :: G' => conj (fun e => conj (proj1 (count_occ_In eq_Probe_dec x (bar A)) (proj1 e))
                                     (proj1 (eval_s1 G') (proj2 e)))
                      (fun s => conj (proj2 (count_occ_In eq_Probe_dec x (bar A)) (proj1 s))
                                     (proj2 (eval_s1 G') (proj2 s)))
  end.