(* ProofTraceToState, corresponding to the second Theorem of Section 3.3,
   maps a BigTrace to a BigStep or a BigStep and ErrorStep
   and maps ErrorTraces to ErrorSteps, defining:
    - BigTrace_to_BigStep
    - ErrorTrace_to_ErrorStep *)



Require Import List.
Require Import SemanticsBase.
Require Import SemanticsState.
Require Import SemanticsTrace.
Require Import ProofPre.
Require Import ProofBase.
Require Import ProofTraceToStateBase.
Require Import ProofTraceToStateInterferance.
Require Import ProofTraceToStateStability.
Require Import ProofTraceToStateMain.



(* ~~~~~~~ VALID ~~~~~~~ *)

Fixpoint Trace_Valid {P P' a} (t : Trace P a P') (v : valid P) : valid P' :=
  match t in Trace P a P' return valid P -> valid P' with
    | SequSkip_T _ => fun v => initial_to_valid (proj2 v)
    | Sequ_T _ _ _ _ t' => fun v => conj (Trace_Valid t' (proj1 v)) (proj2 v)
    | Par_T _ _ _ _ _ _ t0 t1 => fun v => conj (Trace_Valid t0 (proj1 v)) (Trace_Valid t1 (proj2 v))
    | RepIdle_T _ => fun v => v
    | Rep_T _ => fun v => conj (initial_to_valid v) v
    | DetIdle_T gp0 gp1 => fun v => v
    | DetL_T (_, p0) _ => fun v => initial_to_valid (proj1 v)
    | DetR_T _ (_, p1) => fun v => initial_to_valid (proj2 v)
    | Det2_T gp0 gp1 => fun v => v
    | NDetIdle_T gp0 gp1 => fun v => v
    | NDetL_T (_, p0) _ => fun v => initial_to_valid (proj1 v)
    | NDetR_T _ (_, p1) => fun v => initial_to_valid (proj2 v)
    | _ => fun _ => I
  end v.

Fixpoint BigTrace_Valid {P P'} (bt : BigTrace P P') (v : valid P) : valid P' :=
  match bt in BigTrace P P' return valid P -> valid P' with
  | ST _ _ _ t _ _ => fun v => Trace_Valid t v
  | BT _ _ _ bt' bt'' => fun v => BigTrace_Valid bt'' (BigTrace_Valid bt' v)
  end v.



(* ~~~~~~~ COMPLIANT TRACE TO NON-ERRONEOUS OR ERRONEOUS STEP SEQUENCES ~~~~~~~ *)

Fixpoint BigTrace_to_BigStep {P P'} (v : valid P) (bt : BigTrace P P')
                             : sum (BigStep P (Program_to_State P) P' (Program_to_State P'))
                                   {P'' : Program & {s'' : State & prod (BigStep P (Program_to_State P) P'' s'')
                                                                         (ErrorStep P'' s'')}} :=
  match bt in BigTrace P P'
  return valid P -> sum (BigStep P (Program_to_State P) P' (Program_to_State P'))
                        {P'' : Program & {s'' : State & prod (BigStep P (Program_to_State P) P'' s'')
                                                                         (ErrorStep P'' s'')}} with
  | ST _ _ a t c ni => fun v => Trace_to_BSorBSES t v c
  | BT _ _ _ bt' bt'' => fun v => match BigTrace_to_BigStep v bt' with
                                    | inl bs' =>
                                  match BigTrace_to_BigStep (BigTrace_Valid bt' v) bt'' with
                                    | inl bs'' => inl (BS bs' eq_State_refl bs'')
                                    | inr (existT _ (existT _ (bs'', es''))) => inr (existT _ _ (existT _ _
                                      (BS bs' eq_State_refl bs'', es''))) end
                                    | inr bses => inr bses end
  end v.



(* ~~~~~~~ ERRONEOUS TRACES TO ERRONEOUS STEPS ~~~~~~~ *)

Fixpoint detreq_dec (a : Label) : {detreq a} + {~ detreq a} :=
  match a as a return {detreq a} + {~ detreq a} with
    | holds2 G1 G2 => right (fun d => d)
    | ll a0 a1 => match detreq_dec a0, detreq_dec a1 with
                    | left dra0, left dra1 => left (conj dra0 dra1)
                    | right ndra0, _ => right (fun d => match d with conj dra0 _ => ndra0 dra0 end)
                    | _, right ndra1 => right (fun d => match d with conj _ dra1 => ndra1 dra1 end)
                  end
    | _ => left I
  end.

Fixpoint TraceNondet_to_ES' {P P' a l} (t : Trace P a P') (r : req a l) (nd : ~ detreq a) : ErrorStep P l :=
  match t in Trace P a P' return req a l -> ~ detreq a -> ErrorStep P l with
    | Sequ_T _ _ _ _ t0 => fun r => fun nd => Sequ_E _ _ _ (TraceNondet_to_ES' t0 r nd)
    | Par_T _ _ _ _ a0 a1 t0 t1 => fun r => fun nd => match detreq_dec a0, detreq_dec a1 with
                                                        | right ndra0, _ => ParL_E _ _ _ (TraceNondet_to_ES' t0 (proj1 r) ndra0)
                                                        | _, right ndra1 => ParR_E _ _ _ (TraceNondet_to_ES' t1 (proj2 r) ndra1)
                                                        | left dra0, left dra1 => match nd (conj dra0 dra1) with end end
    | Det2_T _ _ => fun r => fun nd => Det2_E _ _ _ (proj1 (eval_s1 _) (proj1 r)) (proj1 (eval_s1 _) (proj2 r))
    | _ => fun _ => fun nd => match nd I with end
  end r nd.

Definition TraceNondet_to_ES {P P' a} (t : Trace P a P') (c : compliant a) (nd : ~ detreq a) : ErrorStep P (pre a) :=
  TraceNondet_to_ES' t c nd.

Definition ErrorTrace_to_ErrorStep {P} (et : ErrorTrace P) : ErrorStep P (Program_to_State P) :=
  match et in ErrorTrace P' with
    | NDR_ET _ _ _ t c nd => eq_rect_r (fun x => ErrorStep _ x) (TraceNondet_to_ES t c nd) (Trace_to_Pre t)
    | Int_ET _ _ _ t c i => eq_rect_r (fun x => ErrorStep _ x) (TraceInt_to_ES t c i) (Trace_to_Pre t)
    | NSt_ET _ _ _ t c ns => eq_rect_r (fun x => ErrorStep _ x) (TraceNotStable_to_ES t c ns) (Trace_to_Pre t)
  end.