(* SemanticsState, corresponding to Figure 5
   and the unchanged rules from Figure 4,
   defines:
    - State
    - SmallStep
    - BigStep
    - Nonstable
    - ErrorStep
    - BigErrorStep *)



Require Import List.
Require Import SemanticsBase.



(* ~~~~~~~ STATE ~~~~~~~ *)

Definition State : Type := list Probe.

Definition eq_State (s1 s2 : State) : Prop :=
  forall p, count_occ eq_Probe_dec s1 p = count_occ eq_Probe_dec s2 p.

Definition state_init : State := nil.

Fixpoint s1 (s : State) (G : Guard) : Prop :=
  match G with
    | nil => True
    | A :: G' => count_occ eq_Probe_dec s (bar A) > 0 /\ s1 s G'
  end.



(* ~~~~~~~ STATE SEMANTICS ~~~~~~~ *)

(* ~~~~~~~ NON-ERRONEOUS STEP ~~~~~~~ *)
Inductive SmallStep : Program -> State -> Program -> State -> Type :=
  | CommB1_S (A : Channel) (s : State)
             (c : ~ In (hat A) s)
      : SmallStep (comm (bang A)) s
                  (comm (bang' A)) ((bar A) :: s)
  | CommB2_S (A : Channel) (s : State)
             (c1 : In (hat A) s)
             (c2 : count_occ eq_Probe_dec s (bar A) = 1)
      : SmallStep (comm (bang' A)) s
                  skip (remove eq_Probe_dec (bar A) s)
  | CommQ1_S (A : Channel) (s : State)
             (c : In (bar A) s)
      : SmallStep (comm (quer A)) s
                  (comm (quer' A)) ((hat A) :: s)
  | CommQ2_S (A : Channel) (s : State)
             (c1 : ~ In (bar A) s)
             (c2 : count_occ eq_Probe_dec s (hat A) = 1)
      : SmallStep (comm (quer' A)) s
                  skip (remove eq_Probe_dec (hat A) s)
  | SequSkip_S (P : Program) (s : State)
      : SmallStep (sequ skip P) s P s
  | Sequ_S (P1 P1' P2 : Program) (s s' : State)
           (ss : SmallStep P1 s P1' s')
      : SmallStep (sequ P1 P2) s (sequ P1' P2) s'
  | ParSkip_S (s : State)
      : SmallStep (par skip skip) s skip s
  | ParL_S (P1 P1' P2 : Program) (s s' : State)
           (ss : SmallStep P1 s P1' s')
      : SmallStep (par P1 P2) s (par P1' P2) s'
  | ParR_S (P1 P2 P2' : Program) (s s' : State)
           (ss : SmallStep P2 s P2' s')
      : SmallStep (par P1 P2) s (par P1 P2') s'
  | Rep_S (P : Program) (s : State)
      : SmallStep (rep P) s (sequ P (rep P)) s
  | DetL_S (GP0 GP1 : Guard * Program) (s : State)
           (c1 : s1 s (fst GP0))
      : SmallStep (det GP0 GP1) s (snd GP0) s
  | DetR_S (GP0 GP1 : Guard * Program) (s : State)
           (c1 : s1 s (fst GP1))
      : SmallStep (det GP0 GP1) s (snd GP1) s
  | NDetL_S (GP0 GP1 : Guard * Program) (s : State)
            (c1 : s1 s (fst GP0))
      : SmallStep (ndet GP0 GP1) s (snd GP0) s
  | NDetR_S (GP0 GP1 : Guard * Program) (s : State)
            (c1 : s1 s (fst GP1))
      : SmallStep (ndet GP0 GP1) s (snd GP1) s.

(* ~~~~~~~ NON-ERRONEOUS STEP SEQUENCE ~~~~~~~ *)
Inductive BigStep : Program -> State -> Program -> State -> Type :=
  | S' {P : Program} {s : State}
       : BigStep P s P s
  | SS {P P' : Program} {s s' : State}
       : SmallStep P s P' s' -> BigStep P s P' s'
  | BS {P P' P'' : Program} {s s' s'' s''' : State}
       : BigStep P s P' s' -> eq_State s' s''
      -> BigStep P' s'' P'' s''' -> BigStep P s P'' s'''.

(* ~~~~~~~ NONSTABLE STEP ~~~~~~~ *)
Inductive Nonstable : Program -> State -> State -> Type :=
  | CommB1_NS (A : Channel) (s s' : State)
              (c1 : ~ In (hat A) s)
              (c2 : In (hat A) s')
      : Nonstable (comm (bang A)) s s'
  | CommB2_NS (A : Channel) (s s' : State)
              (c1 : In (hat A) s)
              (c2 : ~ In (hat A) s')
      : Nonstable (comm (bang' A)) s s'
  | CommQ1_NS (A : Channel) (s s' : State)
              (c1 : In (bar A) s)
              (c2 : ~ In (bar A) s')
      : Nonstable (comm (quer A)) s s'
  | CommQ2_NS (A : Channel) (s s' : State)
              (c1 : ~ In (bar A) s)
              (c2 : In (bar A) s')
      : Nonstable (comm (quer' A)) s s'
  | Sequ_NS (P1 P2 : Program) (s s' : State)
            (ns : Nonstable P1 s s')
      : Nonstable (sequ P1 P2) s s'
  | ParL_NS (P1 P2 : Program) (s s' : State)
            (ns : Nonstable P1 s s')
      : Nonstable (par P1 P2) s s'
  | ParR_NS (P1 P2 : Program) (s s' : State)
            (ns : Nonstable P2 s s')
      : Nonstable (par P1 P2) s s'
  | DetL_NS (GP0 GP1 : Guard * Program) (s s' : State)
            (c1 : s1 s (fst GP0))
            (c2 : ~ s1 s' (fst GP0))
      : Nonstable (det GP0 GP1) s s'
  | DetR_NS (GP0 GP1 : Guard * Program) (s s' : State)
            (c1 : s1 s (fst GP1))
            (c2 : ~ s1 s' (fst GP1))
      : Nonstable (det GP0 GP1) s s'
  | NDetL_NS (GP0 GP1 : Guard * Program) (s s' : State)
             (c1 : s1 s (fst GP0))
             (c2 : ~ s1 s' (fst GP0))
      : Nonstable (ndet GP0 GP1) s s'
  | NDetR_NS (GP0 GP1 : Guard * Program) (s s' : State)
             (c1 : s1 s (fst GP1))
             (c2 : ~ s1 s' (fst GP1))
      : Nonstable (ndet GP0 GP1) s s'.

(* ~~~~~~~ ERRONEOUS STEP ~~~~~~~ *)
Inductive ErrorStep : Program -> State -> Type :=
  | CommB2_E (A : Channel) (s : State)
             (c1 : In (hat A) s)
             (c2 : count_occ eq_Probe_dec s (bar A) > 1)
      : ErrorStep (comm (bang' A)) s
  | CommQ2_E (A : Channel) (s : State)
             (c1 : ~ In (bar A) s)
             (c2 : count_occ eq_Probe_dec s (hat A) > 1)
      : ErrorStep (comm (quer' A)) s
  | Sequ_E (P1 P2 : Program) (s : State)
           (es : ErrorStep P1 s)
      : ErrorStep (sequ P1 P2) s
  | ParL_E (P1 P2 : Program) (s : State)
           (es : ErrorStep P1 s)
      : ErrorStep (par P1 P2) s
  | ParR_E (P1 P2 : Program) (s : State)
           (es : ErrorStep P2 s)
      : ErrorStep (par P1 P2) s
  | Det2_E (GP0 GP1 : Guard * Program) (s : State)
           (c1 : s1 s (fst GP0))
           (c2 : s1 s (fst GP1))
      : ErrorStep (det GP0 GP1) s
  | ParL_NSE (P1 P1' P2 : Program) (s s' : State)
             (ss : SmallStep P1 s P1' s')
             (ns : Nonstable P2 s s')
      : ErrorStep (par P1 P2) s
  | ParR_NSE (P1 P2 P2' : Program) (s s' : State)
             (ss : SmallStep P2 s P2' s')
             (ns : Nonstable P1 s s')
      : ErrorStep (par P1 P2) s.

(* ~~~~~~~ ERRONEOUS STEP SEQUENCE ~~~~~~~ *)
Inductive BigErrorStep : Program -> Type :=
  | BES (P P' : Program) (s' : State)
        (bs : BigStep P state_init P' s')
        (es : ErrorStep P' s')
      : BigErrorStep P.